; compiler.lisp   by Michael Thorpe   2010-08-12

(define (opcode-to-byte opcode)
	(let ((val (get #vm-opcodes opcode)))
		(ifelse val
			val
			(throw 'undefined-opcode opcode))))
(define (byte-to-opcode val)
	(let ((opcode (reverse-get-with-eqv? #vm-opcodes val)))
		(ifelse opcode
			opcode
			(throw 'undefined-opcode-byte val))))
(define (flow-control-opcode? opname)
	(switch opname
		(('goto 'bt 'bf 'catch) #t)
		(else #f)))

(define byte-compile-debug #f)

; Takes an opcode argument and encodes it as a byte stream.  The high bit in
; each byte is set, the second-highest bit in the first byte is sign-extended.
(define (arg-to-raw-bytecode arg)
	(if (or (< arg -0x8000000) (< 0x7FFFFFF arg)) (throw 'opcode-argument-too-big arg))
	(let ((raw (new-string 4))
	      (i 3))
		(loop
			(put-char raw i (bit-select 0x7F 0x80 arg))
			(if (and (<= -64 arg) (<= arg 63))
				(exit-loop (substring raw i (- 4 i))))
			(set i (- i 1))
			(set arg (>> arg 7)))))

(define (byte-compile-newop opname arg)
	(let ((op (new-string 1)))
		(put-char op 0 (opcode-to-byte opname))
		(ifelse (= arg 0)
			op
			(string-concatenate (arg-to-raw-bytecode arg) op))))

; This constructs an instruction that jumps backwards over itself and numbytes
; preceeding bytes.  The return value is just like byte-compile-newop.
(define (byte-compile-reversejump opname numbytes)
	(let ((i 2)
	      (jump #f))
		(loop
			(set jump (byte-compile-newop opname (- 0 i numbytes)))
			(if (= i (string-length jump)) (exit-loop))
			(set i (+ i 1)))
		jump))

; An array where each element is (funcname . handler)
(define #function-compilers #f)
; Takes a function reference (e.g., '#prog2of2) and the handling function
; itself, and adds it to the #function-compilers hash.  The handling function
; will be given the same arguments as byte-compile-helper.  Handling functions,
; as well as byte-compile-helper, return lists of instructions, of the form
; (opcodename . arg) where arg is an integer (except for the localobj opcode
; where arg is an arbitrary object).
(define (function-compiler funcname handler)
	(set #function-compilers (cons (cons funcname handler) #function-compilers))
	#f)

; Numeric comparison opcodes
(define (byte-compile-numeric-comparison code stackitems loopheight arg)
	(if byte-compile-debug (dprint 'byte-compile-numeric-comparison code stackitems loopheight arg))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper (caddr code) (cons #f stackitems) loopheight)
		`((cmp . ,arg))))
(define (byte-compile-lt c s l)
	(byte-compile-numeric-comparison c s l 1))
(function-compiler '< byte-compile-lt)
(define (byte-compile-equal c s l)
	(byte-compile-numeric-comparison c s l 2))
(function-compiler '= byte-compile-equal)
(define (byte-compile-le c s l)
	(byte-compile-numeric-comparison c s l 3))
(function-compiler '<= byte-compile-le)
(define (byte-compile-gt c s l)
	(byte-compile-numeric-comparison c s l 4))
(function-compiler '> byte-compile-gt)
(define (byte-compile-ne c s l)
	(byte-compile-numeric-comparison c s l 5))
(function-compiler '<> byte-compile-ne)
(define (byte-compile-ge c s l)
	(byte-compile-numeric-comparison c s l 6))
(function-compiler '>= byte-compile-ge)

; Basic two-argument opcodes
(define (byte-compile-basic-two-op code stackitems loopheight opcode)
	(if byte-compile-debug (dprint 'byte-compile-basic-two-op code stackitems loopheight opcode))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper (caddr code) (cons #f stackitems) loopheight)
		`((,opcode . 0))))
(define (byte-compile-add c s l)
	(byte-compile-basic-two-op c s l 'add))
(function-compiler '+ byte-compile-add)
(define (byte-compile-sub c s l)
	(byte-compile-basic-two-op c s l 'sub))
(function-compiler '- byte-compile-sub)
(define (byte-compile-mul c s l)
	(byte-compile-basic-two-op c s l 'mul))
(function-compiler '* byte-compile-mul)
(define (byte-compile-div c s l)
	(byte-compile-basic-two-op c s l 'div))
(function-compiler '/ byte-compile-div)
(define (byte-compile-eq? c s l)
	(byte-compile-basic-two-op c s l 'eq))
(function-compiler 'eq? byte-compile-eq?)
(define (byte-compile-cons c s l)
	(byte-compile-basic-two-op c s l 'cons))
(function-compiler 'cons byte-compile-cons)
(define (byte-compile-asr c s l)
	(byte-compile-basic-two-op c s l 'asr))
(function-compiler '>> byte-compile-asr)

; Basic one-argument opcodes
(define (byte-compile-basic-one-op code stackitems loopheight opcode)
	(if byte-compile-debug (dprint 'byte-compile-basic-one-op code stackitems loopheight opcode))
	(if (<> 2 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		`((,opcode . 0))))
(define (byte-compile-car c s l)
	(byte-compile-basic-one-op c s l 'car))
(function-compiler 'car byte-compile-car)
(define (byte-compile-cdr c s l)
	(byte-compile-basic-one-op c s l 'cdr))
(function-compiler 'cdr byte-compile-cdr)
(define (byte-compile-lookup c s l)
	(byte-compile-basic-one-op c s l 'lookup))
(function-compiler 'lookup byte-compile-lookup)
(define (byte-compile-not c s l)
	(byte-compile-basic-one-op c s l 'not))
(function-compiler 'not byte-compile-not)
(define (byte-compile-object-type c s l)
	(byte-compile-basic-one-op c s l 'objtype))
(function-compiler 'object-type byte-compile-object-type)
(define (byte-compile-pair? c s l)
	(byte-compile-basic-one-op c s l 'pair))
(function-compiler 'pair? byte-compile-pair?)
(define (byte-compile-undefine c s l)
	(byte-compile-basic-one-op c s l 'undefine))
(function-compiler 'undefine byte-compile-undefine)

(define (byte-compile-and code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-and code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(let ((b (byte-compile-helper (unquote (caddr code)) stackitems loopheight)))
		(append
			(byte-compile-helper (cadr code) stackitems loopheight)
			`((index . 0))
			`((bf . ,(+ 1 (list-length b))))
			`((pop . 0))
			b)))
(function-compiler '#and byte-compile-and)

(define (byte-compile-append/reversed code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-append/reversed code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper (caddr code) (cons #f stackitems) loopheight)
; If the list to be reversed is empty, skip all the processing
		`((index . 0))
		`((bf . 8))
		`((index . 0))
		`((car . 0))
		`((index . 2))
		`((cons . 0))
		`((put . 1))
		`((cdr . 0))
		`((index . 0))
		`((bt . -8))
; Pop the nil remaining from the reversed list
		`((pop . 0))))
(function-compiler '#append/reversed byte-compile-append/reversed)

(define (byte-compile-bit-select code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-bit-select code stackitems loopheight))
	(if (<> 4 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper (caddr code) (cons #f stackitems) loopheight)
		(byte-compile-helper (cadddr code) (cons #f (cons #f stackitems)) loopheight)
		`((bitsel . 0))))
(function-compiler 'bit-select byte-compile-bit-select)
(define (byte-compile-bit-not code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-bit-not code stackitems loopheight))
	(if (<> 2 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper -1 (cons #f stackitems) loopheight)
		(byte-compile-helper 0 (cons #f (cons #f stackitems)) loopheight)
		`((bitsel . 0))))
(function-compiler 'bit-not byte-compile-bit-not)
(define (byte-compile-bit-and code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-bit-and code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper 0 (cons #f stackitems) loopheight)
		(byte-compile-helper (caddr code) (cons #f (cons #f stackitems)) loopheight)
		`((bitsel . 0))))
(function-compiler 'bit-and byte-compile-bit-and)
(define (byte-compile-bit-or code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-bit-or code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper (caddr code) (cons #f stackitems) loopheight)
		(byte-compile-helper -1 (cons #f (cons #f stackitems)) loopheight)
		`((bitsel . 0))))
(function-compiler 'bit-or byte-compile-bit-or)

(define (byte-compile-bit-xor code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-bit-xor code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper (caddr code) (cons #f stackitems) loopheight)
		`((index . 0))
		`((pushi . -1))
		`((pushi . 0))
		`((bitsel . 0))
		`((bitsel . 0))))
(function-compiler 'bit-xor byte-compile-bit-xor)

(define (byte-compile-catch code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-catch code stackitems loopheight))
	(if (<> 4 (list-length code)) (throw 'argument-error code))
	(let ((type (cadr code))
	      (trialcode (caddr code))
	      (badcode (cadddr code)))
		(if (eq? type '#f)
			(set type #f))
		(set trialcode (byte-compile-helper (unquote trialcode) stackitems loopheight))
		(set badcode (byte-compile-helper (unquote badcode) stackitems loopheight))
		(if type
			(set badcode
				(append
					(byte-compile-helper type stackitems loopheight)
					`((extype . 0))
					`((eq . 0))
					`((bt . 3))
					`((exvalue . 0))
					`((extype . 0))
					`((throw . 0))
					badcode)))
		(append
			`((catch . ,(+ (list-length trialcode) 2)))
			trialcode
			`((toss . 0))
			`((goto . ,(list-length badcode)))
			badcode)))
(function-compiler '#catch byte-compile-catch)

(define (byte-compile-define code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-define code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(ifelse (quoted? (cadr code))
		(let ((var (unquote (cadr code)))
		      (val (byte-compile-helper (caddr code) stackitems loopheight)))
			(unless (eq? 'word (object-type var)) (throw 'argument-error code))
			(let ((pos (list-index stackitems var)))
				(ifelse pos
					(append
						val
						`((index . 0))
						`((put . ,(+ pos 1))))
					(append
						val
						`((localobj . ,var))
						`((define . 0))))))
		(let ((val (byte-compile-helper (caddr code) stackitems loopheight))
		      (var (byte-compile-helper (cadr code) (cons #f stackitems) loopheight)))
			(append
				val
				var
				`((define . 0))))))
(function-compiler '#define byte-compile-define)

; Takes quoted word (replacement variable), filter code, and the list.
; Two loops are run:
;   The first filters the list into a new, reversed list
;   The second reverses the reversed list into the result
(define (byte-compile-filter code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-filter code stackitems loopheight))
	(if (<> 4 (list-length code)) (throw 'argument-error code))
	(unless (quoted? (cadr code)) (throw 'argument-error code))
	(let ((var (unquote (cadr code)))
	      (filter (byte-compile-helper (unquote (caddr code)) (cons var (cons #f (cons #f stackitems))) loopheight)))
		(unless (word? var) (throw 'argument-error code))
		(append
	      (byte-compile-helper (cadddr code) stackitems loopheight)
; If the list is empty, skip all the processing
			`((index . 0))
			`((bf . ,(+ 3 (list-length filter) 20)))
; Start the new list
			`((nil . 0))
; Loop over the filtering code
			`((index . 1))
			`((car . 0))
			filter
			`((put . 0))
			`((bf . 4))
; Add the item to the result list
			`((index . 1))
			`((car . 0))
			`((exch . 0))
			`((cons . 0))
; Move to the next item
			`((index . 1))
			`((cdr . 0))
			`((put . 1))
			`((index . 1))
			`((bt . ,(- 0 2 (list-length filter) 11)))
; Now we have the reversed, filtered list; reverse it
			`((index . 0))
			`((car . 0))
			`((index . 2))
			`((cons . 0))
			`((put . 1))
			`((cdr . 0))
			`((index . 0))
			`((bt . -8))
; Pop the nil remaining from the reversed list
			`((pop . 0)))))
(function-compiler '#filter byte-compile-filter)

(define (byte-compile-ifelse code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-ifelse code stackitems loopheight))
	(if (<> 4 (list-length code)) (throw 'argument-error code))
	(let ((iftrue (byte-compile-helper (unquote (caddr code)) stackitems loopheight))
	      (iffalse (byte-compile-helper (unquote (cadddr code)) stackitems loopheight)))
		(append
			(byte-compile-helper (cadr code) stackitems loopheight)
			`((bf . ,(+ (list-length iftrue) 1)))
			iftrue
			`((goto . ,(list-length iffalse)))
			iffalse)))
(function-compiler '#ifelse byte-compile-ifelse)

(define (byte-compile-let code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-let code stackitems loopheight))
	(if (<> 4 (list-length code)) (throw 'argument-error code))
	(let ((var (unquote (cadr code)))
	      (val (byte-compile-helper (caddr code) stackitems loopheight))
	      (subcode (cadddr code)))
		(unless (eq? 'word (object-type var)) (throw 'argument-error code))
		(append
			val
			(byte-compile-helper (unquote subcode) (cons var stackitems) loopheight)
; "Pop" the var location (which is currently under the topmost value)
			`((put . 0)))))
(function-compiler '#let byte-compile-let)

(define (byte-compile-loop code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-loop code stackitems loopheight))
	(if (<> 2 (list-length code)) (throw 'argument-error code))
	(set code (byte-compile-helper (unquote (cadr code)) stackitems (list-length stackitems)))
	(let ((rev ())
	      (jumpdistance -1))
; Expand next-loop "opcodes"
		(while code
	      (ifelse (and (pair? (car code)) (eq? 'next-loop (caar code)))
				(set rev (cons (cons 'goto jumpdistance) rev))
				(set rev (cons (car code) rev)))
			(set jumpdistance (- jumpdistance 1))
			(set code (cdr code)))
; Do a "next-loop" at the end
		(set code `((pop . 0) (goto . ,(- jumpdistance 1))))
		(set jumpdistance 2)
; Expand exit-loop "opcodes"
		(while rev
	      (ifelse (and (pair? (car rev)) (eq? 'exit-loop (caar rev)))
				(set code (cons (cons 'goto jumpdistance) code))
				(set code (cons (car rev) code)))
			(set jumpdistance (+ jumpdistance 1))
			(set rev (cdr rev))))
	code)
(function-compiler '#loop byte-compile-loop)

(define (byte-compile-exit-loop code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-exit-loop code stackitems loopheight))
	(if (<> 2 (list-length code)) (throw 'argument-error code))
	(let ((extra (- (list-length stackitems) loopheight))
	      (revcode (reverse (byte-compile-helper (cadr code) stackitems loopheight))))
		(if (< extra 0) (throw 'assembler-error code))
		(if (> extra 0)
			(set revcode (cons `(put . ,(- extra 1)) revcode)))
		(if (> extra 1)
			(set revcode (cons `(pop . ,(- extra 2)) revcode)))
		(set revcode (cons `(exit-loop . #f) revcode))
		(reverse revcode)))
(function-compiler '#exit-loop byte-compile-exit-loop)

(define (byte-compile-next-loop code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-next-loop code stackitems loopheight))
	(if (<> 1 (list-length code)) (throw 'argument-error code))
	(let ((extra (- (list-length stackitems) loopheight)))
		(if (< extra 0) (throw 'assembler-error code))
		(set code `((next-loop . #f)))
		(if (> extra 0)
			(set code (cons `(pop . ,(- extra 1)) code)))
		code))
(function-compiler '#next-loop byte-compile-next-loop)

; Takes quoted word (replacement variable), mapper code, and the list.
; Two loops are run:
;   The first maps the list into a new, reversed list
;   The second reverses the reversed list into the result
(define (byte-compile-map code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-map code stackitems loopheight))
	(if (<> 4 (list-length code)) (throw 'argument-error code))
	(unless (quoted? (cadr code)) (throw 'argument-error code))
	(let ((var (unquote (cadr code)))
	      (mapper (byte-compile-helper (unquote (caddr code)) (cons var (cons #f (cons #f stackitems))) loopheight)))
		(unless (word? var) (throw 'argument-error code))
		(append
	      (byte-compile-helper (cadddr code) stackitems loopheight)
; If the list is empty, skip all the processing
			`((index . 0))
			`((bf . ,(+ 3 (list-length mapper) 17)))
; Start the new list
			`((nil . 0))
; Loop over the mapping code
			`((index . 1))
			`((car . 0))
			mapper
			`((put . 0))
			`((exch . 0))
			`((cons . 0))
			`((index . 1))
			`((cdr . 0))
			`((put . 1))
			`((index . 1))
			`((bt . ,(- 0 2 (list-length mapper) 8)))
; Now we have the reversed, mapped list; reverse it
			`((index . 0))
			`((car . 0))
			`((index . 2))
			`((cons . 0))
			`((put . 1))
			`((cdr . 0))
			`((index . 0))
			`((bt . -8))
; Pop the nil remaining from the reversed list
			`((pop . 0)))))
(function-compiler '#map byte-compile-map)
; A version of the above that allows for uses like:
;   (#map-safe 'a '(+ a 10) '(1 2 . 3))
; which will return:
;   (11 12 . 13)
(define (byte-compile-map-safe code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-map-safe code stackitems loopheight))
	(if (<> 4 (list-length code)) (throw 'argument-error code))
	(unless (quoted? (cadr code)) (throw 'argument-error code))
	(let ((var (unquote (cadr code)))
	      (mapper (byte-compile-helper (unquote (caddr code)) (cons var (cons #f (cons #f stackitems))) loopheight)))
		(unless (word? var) (throw 'argument-error code))
		(append
	      (byte-compile-helper (cadddr code) stackitems loopheight)
; If the list is empty, skip all the processing
			`((index . 0))
			`((bf . ,(+ 3 (list-length mapper) 12 (list-length mapper) 11)))
; Start the new list
			`((nil . 0))
; Loop over the mapping code
			`((index . 1))
			`((car . 0))
			mapper
			`((put . 0))
			`((exch . 0))
			`((cons . 0))
			`((index . 1))
			`((cdr . 0))
			`((put . 1))
			`((index . 1))
			`((pair . 0))
			`((bt . ,(- 0 2 (list-length mapper) 9)))
; Check if we reached a non-cons cdr and map it
			`((index . 1))
			`((bf . ,(+ 1 (list-length mapper) 2))) 
			`((index . 1))
			mapper
			`((put . 0))
			`((put . 1))
; Now we have the reversed, mapped list; reverse it
			`((index . 0))
			`((car . 0))
			`((index . 2))
			`((cons . 0))
			`((put . 1))
			`((cdr . 0))
			`((index . 0))
			`((bt . -8))
; Pop the nil remaining from the reversed list
			`((pop . 0)))))
(function-compiler '#map-safe byte-compile-map-safe)

(define (byte-compile-or code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-or code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(let ((b (byte-compile-helper (unquote (caddr code)) stackitems loopheight)))
		(append
			(byte-compile-helper (cadr code) stackitems loopheight)
			`((index . 0))
			`((bt . ,(+ 1 (list-length b))))
			`((pop . 0))
			b)))
(function-compiler '#or byte-compile-or)

(define (byte-compile-prog1of2 code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-prog1of2 code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		(byte-compile-helper (caddr code) (cons #f stackitems) loopheight)
		`((pop . 0))))
(function-compiler '#prog1of2 byte-compile-prog1of2)

(define (byte-compile-prog2of2 code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-prog2of2 code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		`((pop . 0))
		(byte-compile-helper (unquote (caddr code)) stackitems loopheight)))
(function-compiler '#prog2of2 byte-compile-prog2of2)

(define (byte-compile-quote code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-quote code stackitems loopheight))
	(if (<> 2 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		`((quote . 0))))
(function-compiler 'quote byte-compile-quote)

(define (byte-compile-set code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-set code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
	(ifelse (quoted? (cadr code))
		(let ((var (unquote (cadr code)))
		      (val (byte-compile-helper (caddr code) stackitems loopheight)))
			(unless (word? var) (throw 'argument-error code))
			(let ((pos (list-index stackitems var)))
				(ifelse pos
					(append
						val
						`((index . 0))
						`((put . ,(+ pos 1))))
					(append
						val
						`((localobj . ,var))
						`((set . 0))))))
; set is "backwards" (the opcode takes value then var)
		(append
			(byte-compile-helper (caddr code) stackitems loopheight)
			(byte-compile-helper (cadr code) (cons #f stackitems) loopheight)
			`((set . 0)))))
(function-compiler '#set byte-compile-set)

(define (byte-compile-throw code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-throw code stackitems loopheight))
	(if (<> 3 (list-length code)) (throw 'argument-error code))
; throw is "backwards" (the opcode takes value then type)
	(append
		(byte-compile-helper (caddr code) stackitems loopheight)
		(byte-compile-helper (cadr code) (cons #f stackitems) loopheight)
		`((throw . 0))))
(function-compiler 'throw byte-compile-throw)

(define (byte-compile-unquote code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-unquote code stackitems loopheight))
	(if (<> 2 (list-length code)) (throw 'argument-error code))
	(append
		(byte-compile-helper (cadr code) stackitems loopheight)
		`((unquote . 0))))
(function-compiler 'unquote byte-compile-unquote)

(define (byte-compile-helper-unknown-function code stackitems loopheight)
	(if byte-compile-debug (dprint 'byte-compile-helper-unknown-function code stackitems loopheight))
	(let ((args (cdr code))
	      (instructions #f)
	      (numargs (list-length args)))
		(while args
			(set instructions (append instructions (byte-compile-helper (car args) stackitems loopheight)))
			(set args (cdr args))
			(set stackitems (cons #f stackitems)))
		(append
			instructions
			(byte-compile-helper (car code) stackitems loopheight)
			`((call . ,numargs)))))

(define (byte-compile-helper code stackitems loopheight)
	(if (and byte-compile-debug (neq? 'cons (object-type code))) (dprint 'byte-compile-helper code stackitems loopheight))
	(switch (object-type code)
		(('word)
			(let ((pos (list-index stackitems code)))
				(ifelse pos
					`((index . ,pos))
					(switch code
						((#f '#f) `((nil . 0)))
						(('#t) `((nil . 0) (not . 0)))
						(('#exception) `((extype . 0)))
						(('#exceptionvalue) `((exvalue . 0)))
						(else `((localobj . ,code) (lookup . 0)))))))
		(('cons)
			(let ((helper (get #function-compilers (car code))))
				(ifelse helper
					(helper code stackitems loopheight)
					(byte-compile-helper-unknown-function code stackitems loopheight))))
		(('quote)
			(set code (unquote code))
			(ifelse code
				(ifelse (eq? (object-type code) 'integer)
					(byte-compile-helper code stackitems loopheight)
					`((localobj . ,code)))
				`((nil . 0))))
		(('integer)
			(ifelse (and (<= -0x8000000 code) (<= code 0x7FFFFFF))
				`((pushi . ,code))
				`((localobj . ,code))))
		(('nil)
			`((nil . 0)))
		(('bytefunc 'func 'rawfunc 'stream 'string)
			`((localobj . ,code)))
		(else
			(throw 'unknown-objtype code))))

(define (byte-compile-localobj-assign instructions)
	(let ((localobjs #f)
	      (next-index 0))
		(set instructions
			(map i
				(ifelse (and (pair? i) (eq? (car i) 'localobj))
					(let ((pos (get localobjs (cdr i))))
						(unless pos
							(set pos next-index)
							(set localobjs (cons (cons (cdr i) pos) localobjs))
							(set next-index (+ next-index 1)))
						(cons 'localobj pos))
					i)
				instructions))
		(set localobjs (reverse (map i (car i) localobjs)))
		(cons instructions localobjs)))

; Adds up the characters from string[start] to string[end] inclusive.
(define (sum-of-characters string start end)
	(let ((sum 0))
		(while (<= start end)
			(set sum (+ sum (get-char string start)))
			(set start (+ start 1)))
		sum))

(define (byte-compile-flatten instructions)
	(let ((numops (list-length instructions))
	      (minoplen (new-string numops))
	      (maxoplen (new-string numops))
	      (fromto #f)
	      (jumpargs #f)
	      (ops instructions)
	      (i 0)
	      (min 0)
	      (max 0)
	      (op #f))
		(while ops
			(set op (car ops))
			(set ops (cdr ops))
			(ifelse (flow-control-opcode? (car op))
				(progn
					(if (or (< (+ i (cdr op) 1) 0) (< numops (+ i (cdr op) 1)))
						(throw 'assembler-error instructions))
					(ifelse (< (cdr op) 0)
						(set fromto (cons (list op (+ i (cdr op) 1) (- i 1) i) fromto))
						(set fromto (cons (list op (+ i 1) (+ i (cdr op)) i) fromto)))
					(put-char minoplen i 1)
					(put-char maxoplen i 5))
				(progn
					(set op (string-length (byte-compile-newop (car op) (cdr op))))
					(put-char minoplen i op)
					(put-char maxoplen i op)))
			(set i (+ i 1)))
		(while fromto
			(let ((newfromto #f)
			      (progress #f))
				(while fromto
					(set min (sum-of-characters minoplen (cadr (car fromto)) (caddr (car fromto))))
					(set max (sum-of-characters maxoplen (cadr (car fromto)) (caddr (car fromto))))
					(set i (cadddr (car fromto)))
					(if (< (cdar (car fromto)) 0)
						(set min (- 0 (get-char minoplen i) min))
						(set max (- 0 (get-char maxoplen i) max)))
					(ifelse (= min max)
						(progn
							(set op (byte-compile-newop (caar (car fromto)) min))
							(put-char minoplen i (string-length op))
							(put-char maxoplen i (string-length op))
							(set jumpargs (cons (cons (caar fromto) op) jumpargs))
							(set progress #t))
						(progn
							(set min (string-length (byte-compile-newop (caaar fromto) min)))
							(set max (string-length (byte-compile-newop (caaar fromto) max)))
							(if (or (< (get-char minoplen i) min) (> (get-char maxoplen i) max))
								(set progress #t))
							(put-char minoplen i min)
							(put-char maxoplen i max)
							(set newfromto (cons (car fromto) newfromto))))
					(set fromto (cdr fromto)))
				(set fromto newfromto)
				(unless progress
					(exit-loop))))
		(while fromto ; We had jumps we couldn't minimize exactly; set them to min
			(set min (sum-of-characters minoplen (cadr (car fromto)) (caddr (car fromto))))
			(set i (cadddr (car fromto)))
			(if (< (cdar (car fromto)) 0)
				(set min (- 0 (get-char minoplen i) min)))
			(set op (byte-compile-newop (caar (car fromto)) min))
			(if (<> (get-char minoplen i) (string-length op))
				(throw 'assembler-error instructions))
			(set jumpargs (cons (cons (caar fromto) op) jumpargs))
			(set fromto (cdr fromto)))
		(concatenate-strings
			(map op
				(ifelse (flow-control-opcode? (car op))
					(get jumpargs op)
					(byte-compile-newop (car op) (cdr op)))
				instructions))))

(define (byte-assemble args prog)
	(let ((numargs (list-length args)))
		(append
			`((needargs . ,numargs))
			(byte-compile-helper prog (reverse args) #f)
			`((ret . ,numargs)))))

(define (byte-compile args prog)
	(let ((code (byte-assemble args prog)))
; Bytecode optimizers could go here
		(let ((code/objs (byte-compile-localobj-assign code)))
			(#lambda
				(byte-compile-flatten (car code/objs))
				(cdr code/objs)))))

(macro-define (lambda o)
	(list 'byte-compile (quote (cadr o)) (quote (macro-expand (cons 'progn (cddr o))))))

; Trims unused labels and merges series of sequential labels into one label
(define (trim-unused-labels instructions)
	(let ((insts instructions)
	      (used ())
	      (labels ())
	      (replace ())
	      (oldlabel #f))
		(while insts
			(if (and (pair? (car insts)) (flow-control-opcode? (caar insts)))
				(set used (cons (cons (cdar insts) #t) used)))
			(if (word? (car insts))
				(if oldlabel
					(set replace (cons (cons (car insts) oldlabel) replace)))
				(set labels (cons (car insts) labels)))
			(set oldlabel (ifelse (word? (car insts)) (car insts) #f))
			(set insts (cdr insts)))
		(let ((remove ()))
			(while labels
				(if
					(or
						(get replace (car labels))
						(not (get used (car labels))))
					(set remove (cons (car labels) remove)))
				(set labels (cdr labels)))
			(if remove
; insts is already #f
				(while instructions
					(ifelse (and remove (eq? (car instructions) (car remove)))
						(set remove (cdr remove))
						(ifelse
							(and
								replace
								(pair? (car instructions))
								(flow-control-opcode? (caar instructions))
								(get replace (cdar instructions)))
							(set insts (cons (cons (caar instructions) (get replace (cdar instructions))) insts))
							(set insts (cons (car instructions) insts))))
					(set instructions (cdr instructions)))
				(set instructions (reverse insts)))
			instructions)))

(define (op-position-to-word pos)
	(string->symbol (concatenate-strings `("#op-" ,(->string pos)))))

(define (disassemble func)
	(let ((code (code-of-bytefunc func))
	      (objs (objects-of-bytefunc func))
	      (i 0)
	      (c #f)
	      (op #f)
	      (arg #f)
	      (ops ()))
		(while (< i (string-length code))
			(set ops (cons (op-position-to-word i) ops))
			(set c (get-char code i))
			(set arg 0)
			(if (<> 0 (bit-and c 0x80))
				(set arg (ifelse (= 0 (bit-and c 0x40)) 0 -1))
				(while (<> 0 (bit-and c 0x80))
					(set arg (bit-select 0x7F (* arg 128) c))
					(set i (+ i 1))
					(if (= i (string-length code))
						(throw 'disassembly-error func))
					(set c (get-char code i))))
			(set op (byte-to-opcode c))
			(if (flow-control-opcode? op)
				(set arg (op-position-to-word (+ i 1 arg))))
			(if (eq? 'localobj op)
				(set arg (nth-item objs arg)))
			(set ops (cons (cons op arg) ops))
			(set i (+ i 1)))
		(trim-unused-labels (reverse ops))))

(define (disassemble-print file func)
	(let ((ops (disassemble func))
	      (op #f)
	      (arg #f))
		(while ops
			(set op (car ops))
			(set ops (cdr ops))
			(ifelse (word? op)
				(progn
					(write-object file op)
					(write-object file ':))
				(progn
					(set arg (cdr op))
					(set op (car op))
					(write-string file "	")
					(write-object file op)
					(switch op
						(('nil 'pop 'exch 'add 'sub 'mul 'div 'toss 'throw 'define 'lookup 'set 'evaluate 'extype 'exvalue 'eq 'and 'or 'not 'objtype 'cons 'car 'cdr 'quote 'unquote 'bitsel 'asr)
							#f)
						(else
							(write-string file " ")
							(if (and (eq? 'localobj op) (word? arg))
								(write-string file "'"))
							(write-object file arg)))))
			(write-newline file))))

(define (asm-print a)
	(if (eq? 'bytefunc (object-type a))
		(set a (disassemble a)))
	(while a
		(dprint (car a))
		(set a (cdr a))))

(define (evaluate o)
	(switch (object-type o)
		(('nil) #f)
		(('quote) (unquote o))
		(('word) (lookup o))
		(('cons) ((byte-compile #f o)))
		(else o)))

(define (trace expr)
	(let ((old-vm-debug (vm-debug 2)))
		(catch #f
			(set expr (evaluate expr))
			(progn
				(vm-debug old-vm-debug)
				(throw #exception #exceptionvalue)))
		(vm-debug old-vm-debug))
	expr)

(define (compile-file-string outfile var s)
	(write-string outfile "	")
	(write-string outfile var)
	(write-string outfile "=new_string(")
	(write-object outfile s)
	(write-string outfile ",")
	(write-object outfile (string-length s))
	(write-string outfile ");")
	(write-newline outfile)
	(write-string outfile "	if(!")
	(write-string outfile var)
	(write-string outfile ") return(1);")
	(write-newline outfile))

(define (compile-file-word outfile var s)
	(set s (->string s))
	(write-string outfile "	")
	(write-string outfile var)
	(write-string outfile "=new_word(")
	(write-object outfile s)
	(write-string outfile ",")
	(write-object outfile (string-length s))
	(write-string outfile ");")
	(write-newline outfile)
	(write-string outfile "	if(!")
	(write-string outfile var)
	(write-string outfile ") return(1);")
	(write-newline outfile))

(define (compile-file-define-character-macro in outfile c func)
	(unless (eq? (object-type c) 'integer)
		(throw 'compile-file-define-character-macro (list in c func)))
	(write-string outfile "	o=new_object(&integer_objtype);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	o->value.i=")
	(write-object outfile c)
	(write-string outfile ";")
	(write-newline outfile)
	(compile-file-word outfile "s" func)
	(write-string outfile "	s=lookup(s);")
	(write-newline outfile)
	(write-string outfile "	if(!s) return(1);")
	(write-newline outfile)
	(write-string outfile "	c=cons(o,s);")
	(write-newline outfile)
	(write-string outfile "	if(!c) return(1);")
	(write-newline outfile)
	(compile-file-word outfile "s" '#character-macros)
	(write-string outfile "	INCREF(s);")
	(write-newline outfile)
	(write-string outfile "	o=lookup(s);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	o=cons(c,o);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	o=set(s,o);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	DECREF(o);")
	(write-newline outfile))

(define (compile-file-function-compiler in outfile name func)
	(unless (or (word? name) (and (quoted? name) (word? (unquote name))))
		(throw 'compile-file-function-compiler (list in name func)))
	(unless (word? func)
		(throw 'compile-file-function-compiler (list in name func)))
	(ifelse (quoted? name)
		(compile-file-word outfile "s" (unquote name))
		(progn
			(compile-file-word outfile "s" name)
			(write-string outfile "	s=lookup(s);")
			(write-newline outfile)
			(write-string outfile "	if(!s) return(1);")
			(write-newline outfile)))
	(compile-file-word outfile "o" func)
	(write-string outfile "	o=lookup(o);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	c=cons(s,o);")
	(write-newline outfile)
	(write-string outfile "	if(!c) return(1);")
	(write-newline outfile)
	(compile-file-word outfile "s" '#function-compilers)
	(write-string outfile "	INCREF(s);")
	(write-newline outfile)
	(write-string outfile "	o=lookup(s);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	o=cons(c,o);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	o=set(s,o);")
	(write-newline outfile)
	(write-string outfile "	if(!o) return(1);")
	(write-newline outfile)
	(write-string outfile "	DECREF(o);")
	(write-newline outfile))

(define (compile-file in out)
	(let ((infile (open-file in "r"))
	      (outfile (open-file out "w"))
	      (key #f)
	      (o #f)
	      (o2 #f)
	      (s #f)
	      (val #f))
		(loop
			(catch 'end-of-file (set o (read-object infile)) (exit-loop))
			(write-string outfile "// ")
			(write-object outfile o)
			(write-newline outfile)
			(unless (pair? o) (throw 'bad-object o))
			(set o2 (macro-expand o))
			(write-string outfile "// ")
			(write-object outfile o2)
			(write-newline outfile)
			(unless (pair? o2)
				(throw 'compile-file `(in ,1 ,o)))
			(if (eq? (car o2) 'define-character-macro)
				(compile-file-define-character-macro in outfile (cadr o2) (caddr o2))
				(next-loop))
			(if (eq? (car o2) 'function-compiler)
				(compile-file-function-compiler in outfile (cadr o2) (caddr o2))
				(next-loop))
			(unless (or (eq? (car o2) '#define) (eq? (car o2) '#macro-define))
				(throw 'compile-file `(in ,2 ,o)))
			(unless (quoted? (cadr o2))
				(throw 'compile-file `(in ,3 ,o)))
			(set key (unquote (cadr o2)))
			(unless (word? key)
				(throw 'compile-file `(in ,4 ,o ,key)))
			(set val (evaluate (caddr o2)))
			(write-string outfile "// ")
			(write-object outfile val)
			(write-newline outfile)
			(if (eq? 'bytefunc (object-type val))
				(set val (optimize val))
				(write-string outfile "// ")
				(write-object outfile val)
				(write-newline outfile))
			(switch (object-type val)
				(('nil)
					(write-string outfile "	o=INCREF(&nilobj);")
					(write-newline outfile))
				(('bytefunc)
					(let ((revobjs (reverse (objects-of-bytefunc val))))
						(compile-file-string outfile "s" (code-of-bytefunc val))
						(write-string outfile "	c=INCREF(&nilobj);")
						(write-newline outfile)
						(while revobjs
							(let ((arg (car revobjs))
							      (quoteddepth 0))
								(while (quoted? arg)
									(set quoteddepth (+ quoteddepth 1))
									(set arg (unquote arg)))
								(switch (object-type arg)
									(('string)
										(compile-file-string outfile "o" arg))
									(('word)
										(compile-file-word outfile "o" arg))
									(else
										(throw 'compile-file `(in ,5 ,o ,arg))))
								(while (< 0 quoteddepth)
									(write-string outfile "	o=quoteobj(o);")
									(write-newline outfile)
									(write-string outfile "	if(!o) return(1);")
									(write-newline outfile)
									(set quoteddepth (- quoteddepth 1)))
								(write-string outfile "	c=cons(o,c);")
								(write-newline outfile)
								(write-string outfile "	if(!c) return(1);")
								(write-newline outfile))
							(set revobjs (cdr revobjs)))
						(write-string outfile "	o=new_object(&bytefunc_objtype);")
						(write-newline outfile)
						(write-string outfile "	if(!o) return(1);")
						(write-newline outfile)
						(write-string outfile "	o->value.bytefunc.code=s;")
						(write-newline outfile)
						(write-string outfile "	o->value.bytefunc.objs=c;")
						(write-newline outfile)))
				(('cons)
; special support for "unique identifiers" like unquote-identifier and close-parenthesis
					(unless (and (word? (car val)) (word? (cdr val)))
						(throw 'compile-file `(in ,6 ,o ,val)))
					(compile-file-word outfile "o" (car val))
					(compile-file-word outfile "c" (cdr val))
					(write-string outfile "	o=cons(o,c);")
					(write-newline outfile)
					(write-string outfile "	if(!o) return(1);")
					(write-newline outfile))
				(else
					(throw 'compile-file `(in ,7 ,o ,val))))
			(compile-file-word outfile "s" key)
			(switch (car o2)
				(('#define)
					(write-string outfile "	o=define(s,o);")
					(write-newline outfile)
					(write-string outfile "	if(!o) return(1);")
					(write-newline outfile)
					(write-string outfile "	DECREF(o);")
					(write-newline outfile))
				(('#macro-define)
					(write-string outfile "	c=cons(s,o);")
					(write-newline outfile)
					(write-string outfile "	if(!c) return(1);")
					(write-newline outfile)
					(compile-file-word outfile "s" '#macro-expansions)
					(write-string outfile "	INCREF(s);")
					(write-newline outfile)
					(write-string outfile "	o=lookup(s);")
					(write-newline outfile)
					(write-string outfile "	if(!o) return(1);")
					(write-newline outfile)
					(write-string outfile "	o=cons(c,o);")
					(write-newline outfile)
					(write-string outfile "	if(!o) return(1);")
					(write-newline outfile)
					(write-string outfile "	o=set(s,o);")
					(write-newline outfile)
					(write-string outfile "	if(!o) return(1);")
					(write-newline outfile)
					(write-string outfile "	DECREF(o);")
					(write-newline outfile))
				(else
					(throw 'unknown-object-type-in-compile-file o)))
			(write-newline outfile))
		(close-file infile)
		(close-file outfile))
	#f)

